"""
A module containing the Bag class.
"""


class PermissionsError(Exception):
    """
    Base class for policy violation problems.
    """


class ForbiddenError(PermissionsError):
    """
    The provided user cannot do this action.
    """
    pass


class UserRequiredError(PermissionsError):
    """
    To do this action a user is required.
    """
    pass


class Policy(object):
    """
    A container for information about the
    contraints on a bag. A bag is something that
    contains tiddlers. We need to be able to say
    who can do what to do those tiddlers. We also
    need to be able to say who can manage those
    constraints.

    The init parameters represent a default policy.
    The default policy should really come from
    server configuration. Then we can declare this
    installation as open-ish, or closed-ish.
    """

    attributes = ['read', 'write', 'create', 'delete', 'manage', 'accept',
            'owner']

    def __init__(self, owner=None,
            read=None, write=None, create=None, delete=None,
            manage=None, accept=None):
        # avoid "dangerous" warnings from pylint and
        # and possible memory leaks
        if read is None:
            read = []
        if write is None:
            write = []
        if create is None:
            create = []
        if delete is None:
            delete = []
        if manage is None:
            manage = []
        if accept is None:
            accept = []
        self.owner = owner
        self.read = read
        self.write = write
        self.create = create
        self.delete = delete
        self.manage = manage
        self.accept = accept

    def allows(self, user, constraint):
        """
        Is user allowed to perform the action described
        by constraint. The user has a name and some roles,
        either may match in the constraint.
        """
        try:
            roles = user['roles']
        except KeyError:
            roles = []
        user_sign = user['name']

        info_list = self.__getattribute__(constraint)

        if _no_constraint(info_list):
            return True

        user_list = [x for x in info_list if not x.startswith('R:')]

        # always reject if the constraint is NONE
        if _single_value_set(user_list, u'NONE'):
            raise ForbiddenError('%s may not %s' % (user_sign, constraint))

        if _user_valid(user_sign, user_list):
            return True

        role_list = [x[2:] for x in info_list if x.startswith('R:')]

        if _role_valid(roles, role_list):
            return True

        # if the user is set to GUEST (meaning nobody in credentials)
        # then we don't pass, and we need a user
        if user_sign == u'GUEST':
            raise UserRequiredError('real user required to %s' % constraint)

        # we've fallen through, the user we have matches nothing
        raise ForbiddenError('%s may not %s' % (user_sign, constraint))

    def user_perms(self, usersign):
        perms = ['read', 'write', 'create', 'delete']
        matched_perms = []
        for perm in perms:
            try:
                self.allows(usersign, perm)
                matched_perms.append(perm)
            except (UserRequiredError, ForbiddenError):
                pass
        return matched_perms


def create_policy_check(environ, entity, usersign):
    """
    Determine if the user in usersign can create entity
    type.
    """
    try:
        entity_policy = '%s_create_policy' % entity
        policy = environ['tiddlyweb.config'][entity_policy]
    except KeyError:
        raise ForbiddenError('create policy not set for %s' % entity)

    if policy == '':
        return True

    if policy == u'ANY':
        if usersign['name'] != u'GUEST':
            return True
        else:
            raise UserRequiredError('authenticated user required to create')

    if policy == u'ADMIN':
        try:
            if u'ADMIN' in usersign['roles']:
                return True
            else:
                raise ForbiddenError('admin role required to create')
        except KeyError:
            raise ForbiddenError(
                    'admin role required to create, user has no roles')

    raise ForbiddenError('create access denied')


def _single_value_set(target_list, value):
    return len(target_list) == 1 and target_list[0] == value


def _no_constraint(info_list):
    # no constraints then all pass
    try:
        if len(info_list) == 0:
            return True
    except TypeError:
        # constraint not set (weirdness in DB)
        return True

    return False


def _role_valid(roles, role_list):
    # if there is an intersection between the users roles and any roles
    # in the constraint, return true
    if [role for role in roles if role in role_list]:
        return True
    return False


def _user_valid(user_sign, user_list):
    # always allow if the constraint is ANY
    if _single_value_set(user_list, u'ANY'):
        if user_sign != u'GUEST':
            return True

    # if the user is in the constraint list, return true
    if user_sign in user_list:
        return True

    return False



"""
Storage systems for TiddlyWeb.

The base Class and interface for Classes used to get
and put data into a storage system.
"""

from tiddlyweb.store import StoreMethodNotImplemented

TIDDLER_WRITTEN_HANDLERS = []


class StorageInterface(object):
    """
    A Store is a collection of methods that
    either store an object or retrieve an object.

    The interface is fairly simple: For the data
    entities that exist in the TiddlyWeb system there
    (optionally) exists <entity_put> and <entity>_get
    methods in each Store.

    There are also five supporting methods, list_recipes(),
    list_bags(), list_users(), list_bag_tiddlers(), and
    list_tiddler_revisions() that provide methods for
    getting a collection.

    It is useful to understand the classes in the tiddlyweb.model
    package when implementing new StorageInterface classes.
    """

    def __init__(self, store_config=None, environ=None):
        """
        The WSGI environment is made available to the storage system
        so that decisions can be made based on things like the
        request IP. Care should be taken when using environ related
        data as the environment can change from when Store is
        instantiated to when it is actually used. Configuration
        information pulled from the store should only be used
        during the initialization phase.
        """
        if environ is None:
            environ = {}
        if store_config is None:
            store_config = {}
        self.environ = environ
        self.store_config = store_config

    def recipe_delete(self, recipe):
        """
        Remove the recipe from the store, with no impact on the tiddlers.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle deleting recipes')

    def recipe_get(self, recipe):
        """
        Get a recipe from the store, returning a populated recipe object.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle getting recipes')

    def recipe_put(self, recipe):
        """
        Put a recipe into the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle putting recipes')

    def bag_delete(self, bag):
        """
        Remove the bag from the store, including the tiddlers within the bag.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle deleting bags')

    def bag_get(self, bag):
        """
        Get a bag from the store, returning a populated bag object.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle getting bags')

    def bag_put(self, recipe):
        """
        Put a bag into the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle putting bags')

    def tiddler_delete(self, tiddler):
        """
        Delete a tiddler from the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle deleting tiddlers')

    def tiddler_get(self, tiddler):
        """
        Get a tiddler from the store, returning a populated tiddler
        object. tiddler.creator and tiddler.created are based on
        the modifier and modified of the first revision of a tiddler.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle getting tiddlers')

    def tiddler_put(self, tiddler):
        """
        Put a tiddler into the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle putting tiddlers')

    def user_delete(self, user):
        """
        Delete a user from the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle deleting users')

    def user_get(self, user):
        """
        Get a user from the store, returning a populated user object.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle getting users')

    def user_put(self, user):
        """
        Put a user into the store.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle putting users')

    def list_recipes(self):
        """
        Retrieve a list of all recipe objects in the system.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle listing recipes')

    def list_bags(self):
        """
        Retrieve a list of all bag objects in the system.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle listing bags')

    def list_bag_tiddlers(self, bag):
        """
        Retrieve a list of all tiddler objects in the named bag.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle listing bag tiddlers')

    def list_users(self):
        """
        Retrieve a list of all the user objects in the system.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle listing users')

    def list_tiddler_revisions(self, tiddler):
        """
        Retrieve a list of all the revision identifiers
        for one tiddler.
        """
        raise StoreMethodNotImplemented(
                'this store does not handle listing tiddler revisions')

    def tiddler_written(self, tiddler):
        """
        Notify the system that a tiddler has been stored. This is done to
        cause search systems to update, invalidate caches or otherwise deal
        with new content.
        """
        for handler in TIDDLER_WRITTEN_HANDLERS:
            handler(self, tiddler)

    def search(self, search_query):
        """
        Search the entire tiddler store for search_query.
        """
        raise StoreMethodNotImplemented('this store does not provide search')



"""
Send a bag or recipe out HTTP, first serializing to
the correct type.
"""

from tiddlyweb.serializer import Serializer, NoSerializationError
from tiddlyweb.web.http import HTTP415
from tiddlyweb.web.util import get_serialize_type, encode_name
from tiddlyweb.util import sha


def send_entity(environ, start_response, entity):
    """
    Send a bag or recipe out HTTP, first serializing to
    the correct type.
    """
    username = environ['tiddlyweb.usersign']['name']
    try:
        serialize_type, mime_type = get_serialize_type(environ)
        serializer = Serializer(serialize_type, environ)
        serializer.object = entity
        content = serializer.to_string()
    except NoSerializationError:
        raise HTTP415('Content type %s not supported' % mime_type)

    etag_string = '"%s"' % (sha(_entity_etag(entity) +
        encode_name(username) + encode_name(mime_type)).hexdigest())

    start_response("200 OK",
            [('Content-Type', mime_type),
                ('ETag', etag_string),
                ('Vary', 'Accept')])

    if isinstance(content, basestring):
        return [content]
    else:
        return content


def _entity_etag(entity):
    """
    Generate a unique hash from the contents of the entity.
    """
    etag = ''
    etag += entity.name
    for attr in entity.policy.attributes:
        etag += '%s' % getattr(entity.policy, attr)
    try:
        for item in entity:
            etag += '%s' % item
    except TypeError:
        # entity not a iterator
        pass
    return etag.encode('utf-8')



"""
Routines that integrate the basic object classes.

Things like loading up all the tiddlers in a recipe,
listing tiddlers in a bag, and filtering tiddlers.

These are kept in here to avoid a having store
and serialize objects in filters and recipes and the
like.
"""

from tiddlyweb.model.bag import Bag
from tiddlyweb.filters import parse_for_filters, recursive_filter
from tiddlyweb.serializer import TiddlerFormatError
from tiddlyweb.store import NoBagError


def get_tiddlers_from_recipe(recipe, environ=None):
    """
    Return the list of tiddlers that result
    from processing the recipe.

    This list of tiddlers is unique by title with
    tiddlers later in the recipe taking precedence
    over those earlier in the recipe.
    """

    template = recipe_template(environ)
    store = recipe.store
    uniquifier = {}
    for bag, filter_string in recipe.get_recipe(template):
        if isinstance(bag, basestring):
            bag = Bag(name=bag)
        if store:
            bag = store.get(bag)
        for tiddler in filter_tiddlers_from_bag(bag, filter_string, environ=environ):
            uniquifier[tiddler.title] = tiddler
    return uniquifier.values()


def determine_bag_from_recipe(recipe, tiddler, environ=None):
    """
    We have a recipe and a tiddler. We need to
    know the bag in which this tiddler can be found.
    This is different from determine_bag_for_tiddler().
    That one finds the bag the tiddler _could_ be in.
    This is the bag the tiddler _is_ in.

    We reverse the recipe_list, and filter each bag
    according to the rule. Then we look in the list of
    tiddlers and see if ours is in there.
    """
    store = recipe.store
    template = recipe_template(environ)
    for bag, filter_string in reversed(recipe.get_recipe(template)):
        if isinstance(bag, basestring):
            bag = Bag(name=bag)
        if store:
            bag = store.get(bag)
        for candidate_tiddler in filter_tiddlers_from_bag(bag,
                filter_string, environ=environ):
            if tiddler.title == candidate_tiddler.title:
                return bag

    raise NoBagError('no suitable bag for %s' % tiddler.title)


def determine_bag_for_tiddler(recipe, tiddler, environ=None):
    """
    Return the bag which this tiddler would be in if we
    were to save it to the recipe rather than to a default
    bag.

    This is a matter of reversing the recipe list and seeing
    if the tiddler is a part of the bag + filter. If bag+filter
    is true, return that bag.
    """
    template = recipe_template(environ)
    for bag, filter_string in reversed(recipe.get_recipe(template)):
        # ignore the bag and make a new bag
        for candidate_tiddler in filter_tiddlers([tiddler], filter_string, environ=environ):
            if tiddler.title == candidate_tiddler.title:
                if isinstance(bag, basestring):
                    bag = Bag(name=bag)
                return bag

    raise NoBagError('no suitable bag for %s' % tiddler.title)


def get_tiddlers_from_bag(bag):
    """
    Return the list of tiddlers that are in a bag.
    """

    if bag.store:
        for tiddler in bag.store.list_bag_tiddlers(bag):
            if hasattr(tiddler, 'store') and tiddler.store:
                pass
            else:
                try:
                    tiddler = bag.store.get(tiddler)
                except TiddlerFormatError:
                    pass
            yield tiddler
    else:
        for tiddler in bag.tiddlers:
            yield tiddler


def filter_tiddlers(tiddlers, filters, environ=None):
    """
    Return a generator of tiddlers resulting from
    filtering the provided iterator of tiddlers by
    the provided filters.

    If filters is a string, it will be parsed for
    filters.
    """
    if isinstance(filters, basestring):
        filters, _ = parse_for_filters(filters, environ)
    return recursive_filter(filters, tiddlers)


def filter_tiddlers_from_bag(bag, filters, environ=None):
    """
    Return the list of tiddlers resulting from filtering
    bag by filter. The filter is a string that will be
    parsed to a list of filters.
    """
    indexable = bag

    if isinstance(filters, basestring):
        filters, _ = parse_for_filters(filters, environ)
    return recursive_filter(filters, get_tiddlers_from_bag(bag),
            indexable=indexable)


def recipe_template(environ):
    """
    provide a means to specify custom {{ key }} values in recipes
    which are then replaced with the value specified in
    environ['tiddlyweb.recipe_template']
    """
    template = {}
    if environ:
        template = environ.get('tiddlyweb.recipe_template', {})
        try:
            template['user'] = environ['tiddlyweb.usersign']['name']
        except KeyError:
            pass

    return template



"""
Models for TiddlyWeb Entities.

Psuedo-model classes representing the important
entities/data classes in the TiddlyWeb system.

These intentionally are limited, making no effort
to handle their own persistence.
"""



"""
Collection classes.

These classes are used for containing the other
model classes. Because the main reason for having a
collection is to send it out over the web, the
collections keep track of their last-modified time
and generate a hash suitable for use as an ETag.
"""

from tiddlyweb.util import sha


class Collection(object):
    """
    Base class for all collections.

    Can be used directly for random stuff is required.
    """

    def __init__(self):
        self._digest = sha()
        self.modified = 0
        self._container = []

    def __contains__(self, item):
        return item in self._container

    def add(self, thing):
        """
        Add an item to the container, updating
        the digest and modified information.
        """
        self._update_digest(thing)
        self._container.append(thing)
        try:
            modified_string = str(thing.modified)
            modified_string = modified_string.ljust(14, '0')
            if modified_string > self.modified:
                self.modified = modified_string
        except AttributeError:
            pass

    def _update_digest(self, thing):
        """
        Update the digest with this thing.
        """
        self._digest.update(thing)

    def hexdigest(self):
        """
        Return the current hex representation of
        the hash digest of this collection.
        """
        return self._digest.hexdigest()

    def __iter__(self):
        """
        Generate the items in this container.
        """
        for thing in self._container:
            yield thing


class Container(Collection):
    """
    A collection of things which have a name attribute.

    In TiddlyWeb this is for lists of bags and recipes.
    """

    def _update_digest(self, thing):
        """
        Update the digest with this thing.
        """
        self._digest.update(thing.name.encode('utf-8'))


class Tiddlers(Collection):
    """
    A Collection specifically for tiddlers.

    This differs from the base class in how
    the digest is calculated. tiddler.title and
    either tiddler.bag or tiddler.recipe are used.
    """

    def __init__(self):
        Collection.__init__(self)
        self.is_revisions = False
        self.is_search = False

    def _update_digest(self, tiddler):
        """
        Update the digest with information from this tiddler.
        """
        if tiddler.bag:
            container = tiddler.bag
        else:
            container = ''
        self._digest.update(container.encode('utf-8'))
        self._digest.update(tiddler.title.encode('utf-8'))
        self._digest.update(str(tiddler.revision))



"""
Access to Recipe objects via the web. List recipes,
GET a recipe, PUT a recipe, GET the tiddlers
produced by a recipe.
"""

import urllib

from tiddlyweb.model.collections import Tiddlers
from tiddlyweb.model.bag import Bag
from tiddlyweb.model.recipe import Recipe
from tiddlyweb.model.policy import create_policy_check
from tiddlyweb.store import NoRecipeError, NoBagError, \
        StoreMethodNotImplemented
from tiddlyweb.serializer import Serializer, NoSerializationError
from tiddlyweb.web.http import HTTP400, HTTP409, HTTP415, HTTP404
from tiddlyweb.web.sendentity import send_entity
from tiddlyweb.web.sendtiddlers import send_tiddlers
from tiddlyweb.web.listentities import list_entities
from tiddlyweb import control
from tiddlyweb.web import util as web
from tiddlyweb.web.validator import validate_recipe, InvalidBagError


def delete(environ, start_response):
    """
    Delete a recipe, where what delete means
    depends on the store used.
    """
    recipe = _determine_recipe(environ)
    store = environ['tiddlyweb.store']

    recipe.policy.allows(environ['tiddlyweb.usersign'], 'manage')

    try:
        store.delete(recipe)
    except StoreMethodNotImplemented:
        raise HTTP400('Recipe DELETE not supported')

    start_response("204 No Content", [])
    return []


def get(environ, start_response):
    """
    Get the representation of a recipe, based on the
    requested serialization. Will usually show the list
    of bags and filters that make up the recipe.
    """
    recipe = _determine_recipe(environ)
    recipe.policy.allows(environ['tiddlyweb.usersign'], 'read')
    return send_entity(environ, start_response, recipe)


def get_tiddlers(environ, start_response):
    """
    Get the list of tiddlers produced by this
    recipe.
    """
    usersign = environ['tiddlyweb.usersign']
    store = environ['tiddlyweb.store']
    recipe = _determine_recipe(environ)

    # check the recipe can be read
    recipe.policy.allows(usersign, 'read')

    # check the bags in the recipe can be read
    try:
        template = control.recipe_template(environ)
        for bag_name, filter in recipe.get_recipe(template):
            bag = Bag(bag_name)
            bag = store.get(bag)
            bag.policy.allows(usersign, 'read')
    except NoBagError, exc:
        raise HTTP404('recipe %s lists an unknown bag: %s' %
                (recipe.name, exc))

    # from this point forward we know the tiddler are
    # readable

    # get the tiddlers from the recipe and uniquify them
    try:
        candidate_tiddlers = control.get_tiddlers_from_recipe(recipe, environ)
    except NoBagError, exc:
        raise HTTP404('recipe %s lists an unknown bag: %s' %
                (recipe.name, exc))

    tiddlers = Tiddlers()

    for tiddler in candidate_tiddlers:
        tiddler.recipe = recipe.name
        tiddlers.add(tiddler)

    return send_tiddlers(environ, start_response, tiddlers=tiddlers)


def list_recipes(environ, start_response):
    """
    Get a list of all recipes the current user can read.
    """
    store = environ['tiddlyweb.store']
    serialize_type, mime_type = web.get_serialize_type(environ)
    serializer = Serializer(serialize_type, environ)
    return list_entities(environ, start_response, mime_type, store.list_recipes,
            serializer.list_recipes)


def put(environ, start_response):
    """
    Put a new recipe to the server.
    """
    recipe_name = environ['wsgiorg.routing_args'][1]['recipe_name']
    recipe_name = urllib.unquote(recipe_name)
    recipe_name = unicode(recipe_name, 'utf-8')
    recipe_name = web.handle_extension(environ, recipe_name)

    recipe = Recipe(recipe_name)
    store = environ['tiddlyweb.store']
    length = environ['CONTENT_LENGTH']

    usersign = environ['tiddlyweb.usersign']

    try:
        recipe = store.get(recipe)
        recipe.policy.allows(usersign, 'manage')
    except NoRecipeError:
        create_policy_check(environ, 'recipe', usersign)

    try:
        serialize_type = web.get_serialize_type(environ)[0]
    except TypeError:
        raise HTTP400('Content-type header required')

    try:
        serializer = Serializer(serialize_type, environ)
        serializer.object = recipe
        content = environ['wsgi.input'].read(int(length))
        serializer.from_string(content.decode('utf-8'))

        recipe.policy.owner = usersign['name']

        _validate_recipe(environ, recipe)
        store.put(recipe)
    except TypeError, exc:
        raise HTTP400('malformed input: %s' % exc)
    except NoSerializationError:
        raise HTTP415('Content type %s not supported' % serialize_type)

    start_response("204 No Content",
            [('Location', web.recipe_url(environ, recipe))])

    return []


def _validate_recipe(environ, recipe):
    """
    Unless recipe is valid raise a 409 with the reason why.
    """
    try:
        validate_recipe(recipe, environ)
    except InvalidBagError, exc:
        raise HTTP409('Recipe content is invalid: %s' % exc)


def _determine_recipe(environ):
    """
    Interpret URL information to determine the target
    recipe and then get it from the store.
    """
    recipe_name = environ['wsgiorg.routing_args'][1]['recipe_name']
    recipe_name = urllib.unquote(recipe_name)
    recipe_name = unicode(recipe_name, 'utf-8')
    recipe_name = web.handle_extension(environ, recipe_name)
    recipe = Recipe(recipe_name)

    store = environ['tiddlyweb.store']

    try:
        recipe = store.get(recipe)
    except NoRecipeError, exc:
        raise HTTP404('%s not found, %s' % (recipe.name, exc))

    return recipe



"""
Tools for managing command line tools.
"""

import logging
import os
import sys

from tiddlyweb.util import merge_config, std_error_message, initialize_logging


INTERNAL_PLUGINS = ['tiddlyweb.commands']

COMMANDS = {}


def make_command():
    """
    A decorator that marks the decorated method
    as a member of the commands dictionary, with
    associated help.
    """

    def decorate(func):
        """
        Add the function to the commands dictionary.
        """
        func.description = func.__doc__
        COMMANDS[func.__name__] = func
        return func
    return decorate


@make_command()
def usage(args):
    """List this help"""
    if args:
        std_error_message('ERROR: ' + args + '\n')
    for key in sorted(COMMANDS):
        std_error_message('%10s: %s' % (key, COMMANDS[key].description))
    sys.exit(1)


def handle(args):
    """
    Dispatch to the proper function for the command
    given in a args[1].
    """
    from tiddlyweb.config import config
    try:
        if args[1] == '--load':
            args = _external_load(args, config)
    except IndexError:
        args = []

    initialize_logging(config)

    plugins = INTERNAL_PLUGINS
    try:
        plugins.extend(config['twanager_plugins'])
        for plugin in plugins:
            logging.debug('attempting to import twanager plugin %s', plugin)
            # let the import fail with error if it does
            imported_module = __import__(plugin, {}, {}, ['init'])
            imported_module.init(config)
    except KeyError, exc:
        pass # no plugins

    candidate_command = None
    try:
        candidate_command = args[1]
    except IndexError:
        usage('Missing command')

    try:
        args = args[2:]
    except IndexError:
        args = []

    if candidate_command and candidate_command in COMMANDS:
        try:
            logging.debug('running command %s with %s', candidate_command, args)
            COMMANDS[candidate_command](args)
        except IndexError, exc:
            usage('Incorect number of arguments')
        except Exception, exc:
            if config.get('twanager.tracebacks', False):
                raise
            import traceback
            logging.error('twanager error with command "%s %s"\n%s', candidate_command,
                    args, traceback.format_exc())
            usage('%s: %s' % (exc.__class__.__name__, exc.args))
    else:
        usage('No matching command found')


def _external_load(args, config):
    """
    Load a module from by request of the command line.
    """
    module = args[2]
    args = [args[0]] + args[3:]

    if module.endswith('.py'):
        path, module = os.path.split(module)
        module = module.replace('.py', '')
        sys.path.insert(0, path)
        imported_config = _import_module_config(module)
        sys.path.pop(0)
    else:
        imported_config = _import_module_config(module)

    merge_config(config, imported_config)

    return args


def _import_module_config(module):
    """
    Import the module named module to get at its config.
    """
    imported_module = __import__(module, {}, {}, ['config'])
    return imported_module.config



"""
Serialize TiddlyWeb things for the sake of storage and the like.
"""


class TiddlerFormatError(Exception):
    """
    The provided input is insufficient to form a valid Tiddler.
    """
    pass


class NoSerializationError(Exception):
    """
    There is a NoSerialization of this type for the entity.
    """
    pass


class Serializer(object):
    """
    You must set object after initialization.
    """

    def __init__(self, engine, environ=None):
        if environ is None:
            environ = {}
        self.engine = engine
        self.object = None
        self.environ = environ
        self.serialization = None
        self._figure_serialization()

    def _figure_serialization(self):
        """
        Import the required Serialization.
        """
        if self.engine is None:
            raise NoSerializationError
        try:
            imported_module = __import__('tiddlyweb.serializations.%s' % self.engine,
                    {}, {}, ['Serialization'])
        except ImportError, err:
            err1 = err
            try:
                imported_module = __import__(self.engine, {}, {}, ['Serialization'])
            except ImportError, err:
                raise ImportError("couldn't load module for %s: %s, %s" % (self.engine, err, err1))
        self.serialization = imported_module.Serialization(self.environ)

    def __str__(self):
        lower_class = self.object.__class__.__name__.lower()
        try:
            string_func = getattr(self.serialization, '%s_as' % lower_class)
        except AttributeError, exc:
            raise AttributeError('unable to find to string function for %s: %s' % (lower_class, exc))
        return string_func(self.object)

    def to_string(self):
        """
        Provide a string representation of current TiddlyWeb entity object.
        """
        return self.__str__()

    def from_string(self, input_string):
        """
        Turn the provided input_string into a TiddlyWeb entity object of the
        type of self.object. That is: populate self.object based on input_string.
        """
        lower_class = self.object.__class__.__name__.lower()
        try:
            object_func = getattr(self.serialization, 'as_%s' % lower_class)
        except AttributeError, exc:
            raise AttributeError('unable to find from string function for %s: %s' % (lower_class, exc))
        return object_func(self.object, input_string)

    def list_recipes(self, recipes):
        """
        Provide a string representation of the provided recipes
        in the current serialization.
        """
        return self.serialization.list_recipes(recipes)

    def list_bags(self, bags):
        """
        Provide a string representation of the provided bags
        in the current serialization.
        """
        return self.serialization.list_bags(bags)

    def list_tiddlers(self, tiddlers):
        """
        Provide a string representation of the tiddlers in the
        provided Tiddlers collection.
        """
        return self.serialization.list_tiddlers(tiddlers)




import os
from setuptools import setup, find_packages

VERSION = '0.9.4'

setup(
        namespace_packages = ['tiddlywebplugins'],
        name = 'tiddlywebplugins.whoosher',
        version = VERSION,
        description = 'A TiddlyWeb plugin that provides an indexed search interface using Whoosh.',
        long_description=file(os.path.join(os.path.dirname(__file__), 'README')).read(),
        author = 'Chris Dent',
        url = 'http://pypi.python.org/pypi/tiddlywebplugins.whoosher',
        packages = find_packages(exclude=['test']),
        author_email = 'cdent@peermore.com',
        platforms = 'Posix; MacOS X; Windows',
        install_requires = ['setuptools',
            'tiddlyweb>=1.1.0',
            'tiddlywebplugins.utils',
            'Whoosh',
            ],
        zip_safe=False,
        )



"""
Routines related to sending a list of tiddlers out
to the web, including sending those tiddlers and
validating cache headers for list of tiddlers.
"""

from tiddlyweb import control
from tiddlyweb.filters import FilterError, recursive_filter
from tiddlyweb.model.collections import Tiddlers
from tiddlyweb.serializer import Serializer, NoSerializationError
from tiddlyweb.util import sha
from tiddlyweb.web.util import \
        get_serialize_type, http_date_from_timestamp, datetime_from_http_date
from tiddlyweb.web.http import HTTP400, HTTP304, HTTP415


def send_tiddlers(environ, start_response, bag=None, tiddlers=None):
    """
    Output the tiddlers contained in the provided
    bag in a Negotiated representation. Often, but
    not always, a wiki.
    """
    last_modified = None
    etag = None
    download = environ['tiddlyweb.query'].get('download', [None])[0]
    filters = environ['tiddlyweb.filters']

    if tiddlers is None:
        candidate_tiddlers = Tiddlers()
        try:
            for tiddler in control.filter_tiddlers_from_bag(bag, filters):
                candidate_tiddlers.add(tiddler)
        except FilterError, exc:
            raise HTTP400('malformed filter: %s' % exc)
    elif filters:
        candidate_tiddlers = Tiddlers()
        try:
            for tiddler in recursive_filter(filters, tiddlers):
                candidate_tiddlers.add(tiddler)
        except FilterError, exc:
            raise HTTP400('malformed filter: %s' % exc)
    else:
        candidate_tiddlers = tiddlers

    last_modified, etag = _validate_tiddler_list(environ, candidate_tiddlers)

    serialize_type, mime_type = get_serialize_type(environ)

    content_header = ('Content-Type', mime_type)
    cache_header = ('Cache-Control', 'no-cache')
    vary_header = ('Vary', 'Accept')
    response = [content_header, cache_header, vary_header]

    if download:
        response.append(('Content-Disposition',
            'attachment; filename="%s"' % download.encode('utf-8')))

    if last_modified:
        response.append(last_modified)

    if etag:
        response.append(etag)

    try:
        serializer = Serializer(serialize_type, environ)
        output = serializer.list_tiddlers(candidate_tiddlers)
    except NoSerializationError, exc:
        raise HTTP415('Content type not supported: %s:%s, %s' %
                (serialize_type, mime_type, exc))

    start_response("200 OK", response)

    if isinstance(output, basestring):
        return [output]
    else:
        return output


def _validate_tiddler_list(environ, tiddlers):
    """
    Do Etag and Last modified checks on the
    collection of tiddlers.

    If ETag testing is done, no last modified handling
    is done, even if the ETag testing fails.

    If no 304 is raised, then just return last-modified
    and ETag for the caller to use in constructing
    its HTTP response.
    """
    last_modified_number = tiddlers.modified
    last_modified = None
    if last_modified_number:
        last_modified_string = http_date_from_timestamp(last_modified_number)
        last_modified = ('Last-Modified', last_modified_string)

    username = environ.get('tiddlyweb.usersign', {}).get('name', '')

    try:
        _, mime_type = get_serialize_type(environ)
        mime_type = mime_type.split(';', 1)[0].strip()
    except TypeError:
        mime_type = ''
    etag_string = '"%s:%s;%s"' % (tiddlers.hexdigest(),
            str(last_modified_number), sha('%s:%s' %
                (username, mime_type)).hexdigest())
    etag = ('Etag', etag_string)

    incoming_etag = environ.get('HTTP_IF_NONE_MATCH', None)
    if incoming_etag:
        if incoming_etag == etag_string:
            raise HTTP304(incoming_etag)
    else:
        incoming_modified = environ.get('HTTP_IF_MODIFIED_SINCE', None)
        if incoming_modified and \
                (datetime_from_http_date(incoming_modified) >= \
                datetime_from_http_date(last_modified_string)):
            raise HTTP304('')

    return last_modified, etag



"""
HTML based serializers.
"""

import urllib

from tiddlyweb.serializations import SerializationInterface
from tiddlyweb.web.util import encode_name, escape_attribute_value
from tiddlyweb.wikitext import render_wikitext


class Serialization(SerializationInterface):
    """
    Serialize entities and collections to and from
    HTML representations. This is primarily used
    to create browser based presentations.
    """

    def __init__(self, environ=None):
        SerializationInterface.__init__(self, environ)
        self.environ['tiddlyweb.title'] = ''
        self.environ['tiddlyweb.links'] = []

    def list_recipes(self, recipes):
        """
        List the recipes on the system as html.
        """
        self.environ['tiddlyweb.title'] = 'Recipes'
        def wrap_list():
            yield '<ul id="recipes" class="listing">\n'
            for recipe in recipes:
                yield '<li><a href="recipes/%s">%s</a></li>\n' % (
                        encode_name(recipe.name), recipe.name)
            yield '\n</ul>'
        return wrap_list()

    def list_bags(self, bags):
        """
        List the bags on the system as html.
        """
        self.environ['tiddlyweb.title'] = 'Bags'
        def wrap_list():
            yield '<ul id="bags" class="listing">\n'
            for bag in bags:
                yield '<li><a href="bags/%s/tiddlers">%s</a></li>\n' % (
                        encode_name(bag.name), bag.name)
            yield '\n</ul>'
        return wrap_list()

    def list_tiddlers(self, tiddlers):
        """
        List the tiddlers as html.
        """
        server_prefix = self._server_prefix()
        lines = []
        title = 'Tiddlers'
        representation_link = ''
        bag_link = ''
        for tiddler in tiddlers:
            base, base_link, representation_link, title = \
                    self._tiddler_list_info(tiddler)
            if hasattr(tiddlers, 'is_revisions') and tiddlers.is_revisions:
                line = self._tiddler_revision_info(base, base_link, tiddler)
                representation_link += '/%s/revisions' % encode_name(
                        tiddler.title)
                title = 'Revisions of Tiddler %s' % tiddler.title
            else:
                line = self._tiddler_in_bag_info(base, base_link, tiddler)
            lines.append(line)

        if hasattr(tiddlers, 'is_search') and tiddlers.is_search:
            title = 'Found Tiddlers'

        try:
            routing_args = self.environ.get('wsgiorg.routing_args')[1]
        except (TypeError, IndexError, KeyError):
            routing_args = {}

        if 'bag_name' in routing_args and not 'tiddler_name' in routing_args:
            bag_name = routing_args['bag_name']
            bag_name = urllib.unquote(bag_name)
            bag_name = unicode(bag_name, 'utf-8')
            bag_link = ('<div class="baglink"><a href="%s/bags/%s">'
                    'Bag %s</a></div>' % (
                    self._server_prefix(), encode_name(bag_name), bag_name))

        output = "\n".join(lines)
        self.environ['tiddlyweb.title'] = title

        return """
%s
%s
<ul id="tiddlers" class="listing">
%s
</ul>
""" % (self._tiddler_list_header(representation_link), bag_link, output)

    def recipe_as(self, recipe):
        """
        Recipe as html.
        """
        self.environ['tiddlyweb.title'] = 'Recipe %s' % recipe.name
        lines = []
        for bag, filter_string in recipe.get_recipe():
            line = '<li><a href="'
            if not isinstance(bag, basestring):
                bag = bag.name
            line += '%s/bags/%s/tiddlers' % (
                    self._server_prefix(), encode_name(bag))
            if filter_string:
                line += '?%s' % urllib.quote(
                        filter_string.encode('utf-8'), safe=':=;')
            line += '">bag: %s filter:%s</a></li>' % (bag, filter_string)
            lines.append(line)
        output = "\n".join(lines)
        title = 'Bags in Recipe %s' % recipe.name
        tiddler_link = '%s/tiddlers' % encode_name(recipe.name)
        return """
<div class="tiddlerslink"><a href="%s">Tiddlers in Recipe</a></div>
<div id="recipedesc" class="description">%s</div>
<ul id="recipe" class="listing">
%s
</ul>
""" % (tiddler_link, recipe.desc, output)

    def bag_as(self, bag):
        """
        Bag as html.
        """
        self.environ['tiddlyweb.title'] = 'Bag %s' % bag.name
        tiddler_link = '%s/tiddlers' % encode_name(bag.name)
        return """
<div id="bagdesc" class="description">%s</div>
<div class="tiddlerslink"><a href="%s">Tiddlers in Bag %s</a></div>
""" % (bag.desc, tiddler_link, bag.name)

    def tiddler_as(self, tiddler):
        """
        Transform the provided tiddler into an HTML
        representation of the tiddler packaged in a
        DIV. Render the content using the render_wikitext
        subsystem.
        """
        if tiddler.recipe:
            list_link = 'recipes/%s/tiddlers' % encode_name(tiddler.recipe)
            list_title = 'Tiddlers in Recipe %s' % tiddler.recipe
        else:
            list_link = 'bags/%s/tiddlers' % encode_name(tiddler.bag)
            list_title = 'Tiddlers in Bag %s' % tiddler.bag
        list_html = ('<div class="tiddlerslink"><a href="%s/%s" ' %
                (self._server_prefix(), list_link) +
                'title="tiddler list">%s</a></div>' % list_title)
        html = render_wikitext(tiddler, self.environ)
        self.environ['tiddlyweb.title'] = tiddler.title
        return list_html + self._tiddler_div(tiddler) + html + '</div>'

    def _server_prefix(self):
        """
        Return the string that is the server prefix,
        for creating URLs.
        """
        config = self.environ.get('tiddlyweb.config', {})
        return config.get('server_prefix', '')

    def _tiddler_div(self, tiddler):
        """
        The string that starts the div that contains a tiddler.
        """
        return u'<div class="tiddler" title="%s" server.page.revision="%s" ' \
                'modifier="%s" creator="%s" modified="%s" created="%s" tags="%s" %s>' % \
                    (escape_attribute_value(tiddler.title),
                        tiddler.revision,
                        escape_attribute_value(tiddler.modifier),
                        escape_attribute_value(tiddler.creator),
                        tiddler.modified,
                        tiddler.created,
                        escape_attribute_value(self.tags_as(tiddler.tags)),
                        self._tiddler_fields(tiddler.fields))

    def _tiddler_fields(self, fields):
        """
        Turn tiddler fields into a string suitable for
        _tiddler_div.
        """
        output = []
        for key, val in fields.items():
            output.append('%s="%s"' % (key, escape_attribute_value(val)))
        return ' '.join(output)

    def _tiddler_in_bag_info(self, base, base_link, tiddler):
        """
        Get the info for a non-revision tiddler in a list.
        """
        return '<li><a href="%s/%s/%s/tiddlers/%s">%s</a></li>' % (
            self._server_prefix(),
            base,
            base_link,
            encode_name(tiddler.title),
            tiddler.title)

    def _tiddler_list_header(self, representation_link):
        """
        The string we present at the top of a list of tiddlers.
        """
        if representation_link:
            extension_types = self.environ.get('tiddlyweb.config',
                    {}).get('extension_types', {}).keys()
            links = []
            for extension in extension_types:
                link = '<a href="%s.%s">%s</a>' % (representation_link,
                        extension, extension)
                links.append(link)
            link_info = ' '.join(links)
            return """
<div id="tiddlersheader">This list of tiddlers as: %s</div>
""" % (link_info)
        return ''

    def _tiddler_list_info(self, tiddler):
        """
        Get the basic link info needed for listing tiddlers.
        """
        if tiddler.recipe:
            base = 'recipes'
            base_link = encode_name(tiddler.recipe)
            representation_link = '%s/recipes/%s/tiddlers' % (
                    self._server_prefix(), base_link)
            title = 'Tiddlers in Recipe %s' % tiddler.recipe
        else:
            base = 'bags'
            base_link = encode_name(tiddler.bag)
            representation_link = '%s/bags/%s/tiddlers' % (
                    self._server_prefix(), base_link)
            title = 'Tiddlers in Bag %s' % tiddler.bag
        return base, base_link, representation_link, title

    def _tiddler_revision_info(self, base, base_link, tiddler):
        """
        Get the individual revision info for listing revisions.
        """
        return  ('<li><a href="%s/%s/%s/tiddlers/'
            '%s/revisions/%s">%s:%s</a></li>' % (
            self._server_prefix(),
            base,
            base_link,
            encode_name(tiddler.title),
            tiddler.revision,
            tiddler.title,
            tiddler.revision))



"""
A very raw implementation of an openid
challenger that does checkid_setup.

Is OpenID 1.
"""
try:
    from urlparse import parse_qs
except ImportError:
    from cgi import parse_qs

from urlparse import urlparse, urlunparse
import logging
import random
import urllib

import html5lib
from html5lib import treebuilders

from tiddlyweb.web.util import server_base_url, server_host_url, make_cookie
from tiddlyweb.web.challengers import ChallengerInterface

from tiddlyweb.web.http import HTTP302


class Challenger(ChallengerInterface):
    """
    An OpenID challenger that asks the client for
    an openid and then follows the checkid_setup process
    to validate the ID. Once valid, a cookie is given
    to the browser which is later used by the simple_cookie
    credential extractor.
    """

    def challenge_get(self, environ, start_response):
        """
        Respond to a GET request. If there is no openid_mode,
        send a form to ask for an openid. If openid_mode was
        cancel, then the client cancelled the process somewhere
        in transit, in which case send the form again asking for
        another ID. If mod is id_res, then we have previously
        sent content a request to the OpenID server and now must
        handle the response.
        """
        redirect = (environ['tiddlyweb.query'].
                get('tiddlyweb_redirect', ['/'])[0])
        openid_mode = environ['tiddlyweb.query'].get('openid.mode', [None])[0]

        if openid_mode:
            if openid_mode == 'id_res':
                return self._handle_server_response(environ, start_response)
            if openid_mode == 'cancel':
                return self._send_openid_form(
                        environ, start_response, redirect,
                        message='Try a different OpenID?')
        return self._send_openid_form(environ, start_response, redirect)

    def challenge_post(self, environ, start_response):
        """
        Respond to a POST request. If the form has been filled out, we need
        to find the server and send the user to talk to it. Otherwise, ask
        the user to fill out the form again.
        """
        redirect = (environ['tiddlyweb.query'].
                get('tiddlyweb_redirect', ['/'])[0])
        openid = environ['tiddlyweb.query'].get('openid', [''])[0]

        if len(openid):
            return self._find_speak_to_server(environ, start_response,
                    redirect, openid)
        else:
            return self._send_openid_form(environ, start_response, redirect,
                    message='Enter an OpenID')

    def _handle_server_response(self, environ, start_response):
        """
        The user has gone to their OpenID server and confirmed we can
        have accesss. We need to send a signed copy of the data we've
        recieved in a second request to the server, to confirm. If the
        response is 'is_valid: true' then we passed and we can send a
        cookie to the user.
        """
        request_info = environ['tiddlyweb.query']
        parsed_return_to = parse_qs(request_info['openid.return_to'][0])
        openid_server = parsed_return_to['openid_server'][0]
        redirect = parsed_return_to['tiddlyweb_redirect'][0]
        data = {
            'openid.mode': 'check_authentication',
            'openid.sig': request_info['openid.sig'][0],
            'openid.signed': request_info['openid.signed'][0],
            'openid.assoc_handle': request_info['openid.assoc_handle'][0],
            }
        for item in request_info['openid.signed'][0].split(','):
            if item in ['mode', 'sig', 'signed', 'assoc_handle']:
                continue
            key = 'openid.%s' % item
            try:
                data[key] = request_info[key][0]
            except KeyError:
                pass # signed key not in request
        post_data = urllib.urlencode(data)

        response = urllib.urlopen(openid_server, post_data).read()

        if 'is_valid:true' in response:
            return self._respond_success(parsed_return_to, redirect, environ,
                    start_response)

        return self._send_openid_form(environ, start_response, redirect,
                message=response)

    def _respond_success(self, parsed_return_to, redirect, environ,
            start_response):
        """
        If the openid server validates our key checking, then
        set the cookie and redirect the user.
        """
        usersign = parsed_return_to['usersign'][0]
        if 'http' in usersign:
            usersign = usersign.split('://', 1)[1]
        uri = '%s%s' % (server_host_url(environ), redirect)
        secret = environ['tiddlyweb.config']['secret']
        cookie_age = environ['tiddlyweb.config'].get('cookie_age', None)
        cookie_header_string = make_cookie('tiddlyweb_user', usersign,
                mac_key=secret, path=self._cookie_path(environ),
                expires=cookie_age)
        logging.debug('303 to %s', uri)
        start_response('303 See Other',
                [('Location', uri.encode('utf-8')),
                ('Content-Type', 'text/plain'),
                ('Set-Cookie', cookie_header_string)])
        return [uri]

    def _send_openid_form(self, environ, start_response, redirect,
            status='401 Unauthorized', message=''):
        """
        Send a form requesting an openid to the client.
        """
        start_response(status, [('Content-Type', 'text/html')])
        environ['tiddlyweb.title'] = 'OpenID Login'
        return [
"""
<pre>
%s
<form action="" method="POST">
OpenID: <input name="openid" size="60" />
<input type="hidden" name="tiddlyweb_redirect" value="%s" />
<input type="submit" value="submit" />
</form>
</pre>
""" % (message, redirect)]

    def _find_speak_to_server(self, environ, start_response, redirect, openid):
        """
        Search on the users provided OpenID page for the openid.server
        or openid.delegate and redirect the user there with proper
        parameters.
        """
        if not (openid.startswith('http://') or
                openid.startswith('https://')):
            openid = 'http://%s' % openid

        if openid.endswith('/'):
            openid = openid.rstrip('/')

        try:
            htmlpage = urllib.urlopen(openid).read()
        except IOError, exc:
            return self._send_openid_form(
                    environ, start_response, redirect,
                    message='Unable to talk to opendid server: %s' % exc)

        parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder('dom'))
        dom = parser.parse(htmlpage)

        # Save aside the original open id, in case we'll be doing a
        # delegation.
        original_openid = openid

        # XXX This next bit of complexity brought to you by the letters
        # D, O and M.
        links = dom.getElementsByTagName('link')
        link = None
        for candidate_link in links:
            attributes = candidate_link.attributes
            if (attributes.has_key('rel') and
                    attributes.getNamedItem('rel').value == 'openid.server'):
                link = attributes.getNamedItem('href').value
            if (attributes.has_key('rel') and
                    attributes.getNamedItem('rel').value == 'openid.delegate'):
                openid = attributes.getNamedItem('href').value
        if not link:
            return self._send_openid_form(
                    environ, start_response, redirect,
                    message='Unable to find openid server')

        # parse the openid.server link to separate out any possible query
        # string
        scheme, netloc, path, params, link_query, frag = urlparse(link)
        link = urlunparse((scheme, netloc, path, params, '', ''))
        if link_query:
            link_query = link_query + '&'
        request_uri = ('%s?%sopenid.mode=checkid_setup&openid.identity=%s'
                '&openid.return_to=%s' % (link, link_query, urllib.quote(openid, safe=''),
                    urllib.quote(self._return_to(environ, redirect, link,
                    original_openid), safe='')))

        logging.debug('302 to %s', request_uri)
        raise HTTP302(request_uri)

    def _return_to(self, environ, redirect, link, usersign):
        """
        Generate the URL to which the user should return to after
        visiting their OpenID server.
        """
        return ('%s/challenge/openid?nonce=%s&tiddlyweb_redirect=%s'
                '&openid_server=%s&usersign=%s' % (
                    server_base_url(environ), self._nonce(),
                    redirect, link, usersign))

    def _nonce(self):
        """
        Generate a random string.
        """
        return ''.join([random.choice('ABCDEFGHIJHKLMNOPQRSTUVWXYZ') for
            x in xrange(8)])



"""
A module containing the Tiddler class and related functions.
"""

import re

from datetime import datetime


def current_timestring():
    """
    Translate (now) into a TiddlyWiki conformat timestring.
    """
    time_object = datetime.utcnow()
    return unicode(time_object.strftime('%Y%m%d%H%M%S'))


def string_to_tags_list(string):
    """
    Given a string representing tags in TiddlyWiki
    format parse them into a list of tag strings.
    """
    tags = []
    tag_matcher = re.compile(r'([^ \]\[]+)|(?:\[\[([^\]]+)\]\])')
    for match in tag_matcher.finditer(string):
        if match.group(2):
            tags.append(match.group(2))
        elif match.group(1):
            tags.append(match.group(1))

    return tags


class Tiddler(object):
    """
    The universal content object in the TiddlyWiki
    universe, corresponding to a Page in other wiki
    systems. A Tiddler has text and some associated
    metadata. The text can be anything, but is usually
    wikitext in some form, or Javascript code to be
    used as a plugin.

    Tiddler is intentional just a container of data.
    That is, it has no methods which change the state
    of attributes in the Tiddler or manipulate the tiddler.
    Changing the attributes is done by directly changing
    the attributes. This is done to make the Tiddler
    easier to store and serialize in a diversity of ways.

    A Tiddler has several attributes:

    title: The name of the tiddler. Required.
    created: A string representing when this tiddler was
            created.
    modified: A string representing when this tiddler was
             last changed. Defaults to now.
    modifier: A string representing a personage that changed
              this tiddler in some way. This doesn't necessarily
              have any assocation with the tiddlyweb.usersign,
              though it may.
    tags: A list of strings that describe the tiddler.
    fields: An arbitrary dictionary of extended fields on the tiddler.
    text: The contents of the tiddler. A string.
    revision: The revision of this tiddler. An int.
    bag: The name of the bag in which this tiddler exists,
         if any. Usually set by internal code.
    recipe: The name of the recipe in which this tiddler exists,
            if any. Usually set by internal code.
    store: A reference to the Store object which retrieved
           this tiddler from persistent storage.
    """

    slots = ['title',
            'created',
            'modified',
            'modifier',
            'creator',
            'tags',
            'fields',
            'text',
            'type',
            'revision',
            'bag',
            'recipe',
            'store']

    def __init__(self, title=None, bag=None):
        """
        Create a new Tiddler object.

        A title is required to ceate a tiddler.
        """
        if title:
            title = unicode(title)
        if bag:
            bag = unicode(bag)
        self.title = title
        self.bag = bag

        self._creator = ''
        self.modifier = None
        self.created = ''
        self.modified = current_timestring()
        self.tags = []
        self.fields = {}
        self.text = ''
        self.type = None
        self.revision = 0
        self.recipe = None
        # reference to the store which 'got' us
        self.store = None

    def _get_creator(self):
        if not self._creator:
            self._creator = self.modifier
        return self._creator

    def _set_creator(self, creator):
        self._creator = creator

    creator = property(_get_creator, _set_creator)

    def __repr__(self):
        """
        Make the printed tiddler include the title so we
        can distinguish between them while debugging.
        """
        return self.title + object.__repr__(self)



"""
Make a query into the store to find some
tiddlers and list them in the interface.
"""

import logging

import urllib

from tiddlyweb.web.http import HTTP400
from tiddlyweb.model.collections import Tiddlers
from tiddlyweb.model.bag import Bag
from tiddlyweb.model.policy import ForbiddenError, UserRequiredError
from tiddlyweb.store import StoreMethodNotImplemented
from tiddlyweb.web.sendtiddlers import send_tiddlers


def get_search_query(environ):
    """
    Inspect tiddlyweb.query in the environment to get
    the search query.
    """
    try:
        search_query = environ['tiddlyweb.query']['q'][0]
        search_query = urllib.unquote(search_query)
    except (KeyError, IndexError):
        raise HTTP400('query string required')
    return search_query


def get_tiddlers(environ):
    """
    Call search in the store with search query to
    get the generator of tiddlers matching the
    query.
    """
    search_query = get_search_query(environ)
    store = environ['tiddlyweb.store']
    try:
        tiddlers = store.search(search_query)
        logging.debug('got search results from store')
    except StoreMethodNotImplemented:
        raise HTTP400('Search system not implemented')
    return tiddlers


def get(environ, start_response):
    """
    Perform a search on the store. What search
    means and what results are returned is dependent
    on the search implementation (if any) in the
    chosen store.
    """
    store = environ['tiddlyweb.store']

    tiddlers = get_tiddlers(environ)

    usersign = environ['tiddlyweb.usersign']

    candidate_tiddlers = Tiddlers()
    candidate_tiddlers.searchbag = True

    bag_readable = {}

    for tiddler in tiddlers:
        if not (hasattr(tiddler, 'store') and tiddler.store):
            tiddler = store.get(tiddler)
        try:
            if bag_readable[tiddler.bag]:
                candidate_tiddlers.add(tiddler)
        except KeyError:
            bag = Bag(tiddler.bag)
            bag = store.get(bag)
            try:
                bag.policy.allows(usersign, 'read')
                candidate_tiddlers.add(tiddler)
                bag_readable[tiddler.bag] = True
            except(ForbiddenError, UserRequiredError):
                bag_readable[tiddler.bag] = False

    return send_tiddlers(environ, start_response, tiddlers=candidate_tiddlers)
